#if !defined(lint)
static char copyright[] = "Copyright (c) 1994 Technical University of Chemnitz-Zwickau";
#endif

/*
 * web500gw.c, Version 1.4: 
 *
 *               A LDAP based WWW - X.500 gateway
 *
 * written by: Frank Richter, Frank.Richter@hrz.tu-chemnitz.de
 *             Technical University of Chemnitz-Zwickau, Germany, 1994
 *
 * Copyright (c) 1994 Technical University of Chemnitz-Zwickau.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the Technical University of Chemnitz-Zwickau. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * ----------------------------------------------------------------------
 * This source is based on go500gw by Tim Howes, University of Michigan:
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 * ----------------------------------------------------------------------
 */

/* For changes see file CHANGES */

#include "lber.h"
#include "ldap.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifdef __hpux
#include <syslog.h>
#else
#include <sys/syslog.h>
#endif
#include <sys/wait.h>
#include <signal.h>

/*
 *******************************************************************************
 * If you are not a TU Chemnitz site, you probably want to tailor the following:
 *******************************************************************************
 */

#ifndef WEB500DN
#define WEB500DN	"cn=WWWeb/X.500-Gateway,cn=saturn,ou=Rechenzentrum,o=Technische Universitaet Chemnitz-Zwickau,c=DE"
#endif

#ifdef BASEDN
#ifndef DEFAULT_BASEDN
#define DEFAULT_BASEDN	"o=Technische Universitaet Chemnitz-Zwickau, c=DE"
#endif
#endif

#ifndef	ROOT
#define	ROOT		"The World"
#endif	ROOT

#ifndef HELPFILE
#define HELPFILE	"/usr/local/etc/web500gw.help"
#endif

#ifndef LDAPHOST
#define LDAPHOST	"localhost"
#endif

#ifndef DEFAULT_FILTERFILE
#define DEFAULT_FILTERFILE	"/usr/local/etc/ldapfilter.conf"
#endif

#ifndef DEFAULT_FRIENDLYFILE
#define DEFAULT_FRIENDLYFILE	"/usr/local/etc/ldapfriendly"
#endif

#ifndef G3TOXBM
#define	G3TOXBM		"/usr/local/bin/g3toxbm"
#endif

#ifndef JPEGTOGIF
#define	JPEGTOGIF	"/uni/global/bin/djpeg -gif"
#endif

#ifdef MAIL_FINGER
#ifdef	RUTGERS
#define	FINGER_CMD	"http://www-ns.rutgers.edu/htbin/finger?entry="
#endif
#endif

#ifndef DEFAULT_TOP_OF_BODY_FILE
#define DEFAULT_TOP_OF_BODY_FILE	"/usr/local/etc/web500.head"
#endif
#ifndef DEFAULT_END_OF_BODY_FILE
#define DEFAULT_END_OF_BODY_FILE	"/usr/local/etc/web500.foot"
#endif

/*
 *************************************************************************
 * The rest of this stuff probably does not need to be changed
 *************************************************************************
 */

/* HREFDN is still not used */
#ifdef HREFDN
#define hrefdn(dn, action)	href_dn(dn, action)
#else
#define hrefdn(dn, action)	dn
#endif

#define TIMEOUT		240
#define WEB500PORT	8888

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#ifdef __hpux
#define getdtablesize()	_NFILE
#endif

int	debug;
int	dosyslog;
int	inetd;

char		*ldaphost = LDAPHOST;
int		ldapport = LDAP_PORT;
int		searchaliases = 1;
char		*helpfile = HELPFILE;
char		*filterfile = DEFAULT_FILTERFILE;
char		*friendlyfile = DEFAULT_FRIENDLYFILE;
char		*friendlyDesc = NULL;
FriendlyMap     *fm = NULL;
LDAPFiltDesc	*filtd;
#ifdef BASEDN
char		*basedn = DEFAULT_BASEDN;
#else
char		*basedn = "";
#endif

#define from_hex(c)	((c>='0')&&(c<='9') ? c-'0' : (c>='A')&&(c<='F') ?\
			c-'A'+10 : (c>='a')&&(c<='f') ? c-'a'+10 : 0)

static set_socket();
static wait4child();
static do_queries();
static do_menu();
static do_search();
static do_read();
static do_help();
static do_sizelimit();
static do_timelimit();
static do_error();
static do_audio();
static do_pict();
static do_form();
static do_modify();
static time_t gtime();
static make_scope();
static explain_error();
static sort_result();
static char	*format_date();
static char	*friendly_dn();
static char	*clean_ufn();
#ifdef DEREFALIAS
static void	deref();
#endif

char *hex_encode();
char *hex_decode();
char *hex_qdecode();
char *html_encode();
char *form_encode();
char *href_dn();
extern char	version[];

int	http = 0;	/* HTTP Version ??? */

/*
 * HTTP request we are implementing
 */

#define	UNKNOWN	0
#define GET	1
#define HEAD	2
int	request = UNKNOWN;

/*
 * HTTP response status
 */
#define DOCUMENT_FOLLOWS 200
#define REDIRECT 302
#define BAD_REQUEST 400
#define AUTH_REQUIRED 401
#define FORBIDDEN 403
#define NOT_FOUND 404
#define SERVER_ERROR 500
#define NOT_IMPLEMENTED 501

#define PRINT_HEADER 	(fprintf(fp, "HTTP/1.0 %d OK\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: text/html\r\n\r\n", DOCUMENT_FOLLOWS, version))


typedef struct attr {
	char	*attr_name;
	int	 attr_numb;	/* 0 - don't allow modify, just display value */
} attr;

/* attributes we let to modify */
struct attr mod_attr[24] = { "cn", 0, "title", 1, "telephoneNumber", 1, 
		"facsimileTelephoneNumber", 0, "postalAddress", 0, "mail", 1, 
		"description", 3, 
		"info", 2, "roomNumber", 1, "userClass", 1, "uid", 2, 
		"homePostalAddress", 1, "homePhone", 1, "drink", 1, NULL};

/* objectclasses we let to modify */
char	*mod_oc[]	= { "person", NULL};

#define MAXCOUNT 300
struct dncompare {
	char *string;
	char *href;
} *dnlist[MAXCOUNT];

static int compare(a,b)
struct dncompare **a, **b;
{
	return strcmp((*a)->string,(*b)->string);
}

/* Flags for print_attr */
#define	DEFAULT		0
#define	MULTILINE	1
#define	HREF		2
#define	FINGER		3
#define	DATE		4
#define	URL		5


static usage( name )
char	*name;
{
	fprintf( stderr, "usage: %s [-d debuglevel] [-p port] [-l] [-x ldaphost] [-a] [-h helpfile] [-f filterfile] [-b basedn]\n" );
	exit( 1 );
}

main (argc, argv)
int	argc;
char	**argv;
{
	int			s, ns, rc;
	int			port = -1;
	int			tblsize;
	int			i, pid;
	char			*myname;
	fd_set			readfds;
	struct hostent		*hp;
	struct sockaddr_in	from;
	int			fromlen;
	int			wait4child();
	extern char		*optarg;
	extern char		**Argv;
	extern int		Argc;

	/* for setproctitle */
        Argv = argv;
        Argc = argc;

	while ( (i = getopt( argc, argv, "P:ad:f:h:lp:x:Ib:" )) != EOF ) {
		switch( i ) {
		case 'a':
			searchaliases = 0;
			break;
		case 'b':
			basedn = strdup( optarg );
			break;
		case 'd':
			debug = atoi( optarg );
#ifdef LDAP_DEBUG
			ldap_debug = debug;
#else
			fprintf( stderr, "warning: ldap debugging requires LDAP_DEBUG\n" );
#endif
			break;

		case 'f':
			filterfile = strdup( optarg );
			break;

		case 'h':
			helpfile = strdup( optarg );
			break;

		case 'l':
			dosyslog = 1;
			break;

		case 'p':
			port = atoi( optarg );
			break;

		case 'P':
			ldapport = atoi( optarg );
			break;

		case 'x':
			ldaphost = strdup( optarg );
			break;
		case 'I':
			inetd = 1;
			break;

		default:
			usage( argv[0] );
		}
	}

	/* detach if stderr is redirected or no debugging */
	if ( inetd == 0 )
		(void) detach( debug );

	if ( (myname = strrchr( argv[0], '/' )) == NULL )
		myname = strdup( argv[0] );
	else
		myname = strdup( myname + 1 );

	if ( dosyslog && openlog( myname, LOG_PID | LOG_NOWAIT, LOG_LOCAL3 )
	    < 0 ) {
		perror("openlog");
	}
	if ( dosyslog )
		syslog( LOG_INFO, "initializing" );

	/* set up the socket to listen on */
	if ( inetd == 0 ) {
		s = set_socket( port );

		/* arrange to reap children */
		(void) signal( SIGCHLD, wait4child );
	}

	if ( (filtd = ldap_init_getfilter( filterfile )) == NULL ) {
		fprintf( stderr, "Cannot open filter file (%s)\n", filterfile );
		exit( 1 );
	}

	if ( inetd ) {
		fromlen = sizeof(from);
		if ( getpeername( 0, (struct sockaddr *) &from, &fromlen )
		    == 0 ) {
			hp = gethostbyaddr( (char *) &(from.sin_addr.s_addr),
			    sizeof(from.sin_addr.s_addr), AF_INET );
			if (debug) {
			    fprintf( stderr, "connection from %s (%s)\n",
			    (hp == NULL) ? "unknown" : hp->h_name,
			    inet_ntoa( from.sin_addr ), 0 );
			}
			if ( dosyslog ) {
				syslog( LOG_INFO, "connection from %s (%s)",
				    (hp == NULL) ? "unknown" : hp->h_name,
				    inet_ntoa( from.sin_addr ) );
			}

			setproctitle( hp == NULL ? inet_ntoa( from.sin_addr ) :
			    hp->h_name );
		}

		do_queries( 0 );

		close( 0 );

		exit( 0 );
	}

	tblsize = getdtablesize();
	for ( ;; ) {
		FD_ZERO( &readfds );
		FD_SET( s, &readfds );

		if ( (rc = select( tblsize, &readfds, 0, 0 ,0 )) == -1 ) {
			if ( debug ) perror( "select" );
			continue;
		} else if ( rc == 0 ) {
			continue;
		}

		if ( ! FD_ISSET( s, &readfds ) )
			continue;

		fromlen = sizeof(from);
		if ( (ns = accept( s, &from, &fromlen )) == -1 ) {
			if ( debug ) perror( "accept" );
			exit( 1 );
		}

		hp = gethostbyaddr( (char *) &(from.sin_addr.s_addr),
		    sizeof(from.sin_addr.s_addr), AF_INET );

		if ( dosyslog ) {
			syslog( LOG_INFO, "TCP connection from %s (%s)",
			    (hp == NULL) ? "unknown" : hp->h_name,
			    inet_ntoa( from.sin_addr ) );
		}
		if ( debug ) {
                            fprintf( stderr, "connection from %s (%s)\n",    
                            (hp == NULL) ? "unknown" : hp->h_name,
                            inet_ntoa( from.sin_addr ), 0 );
		}
	

		switch( pid = fork() ) {
		case 0:		/* child */
			close( s );
			do_queries( ns );
			break;

		case -1:	/* failed */
			perror( "fork" );
			break;

		default:	/* parent */
			close( ns );
			if ( debug )
				fprintf( stderr, "forked child %d\n", pid );
			break;
		}
	}
	/* NOT REACHED */
}

static set_socket( port )
int	port;
{
	int			s, one;
	struct sockaddr_in	addr;

	if ( port == -1 )
		port = WEB500PORT;

	if ( (s = socket( AF_INET, SOCK_STREAM, 0 )) == -1 ) {
                perror( "socket" );
                exit( 1 );
        }

        /* set option so clients can't keep us from coming back up */
	one = 1;
        if ( setsockopt( s, SOL_SOCKET, SO_REUSEADDR, (char *) &one,
	    sizeof(one) ) < 0 ) {
                perror( "setsockopt" );
                exit( 1 );
        }

        /* bind to a name */
        addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = INADDR_ANY;
        addr.sin_port = htons( port );
        if ( bind( s, &addr, sizeof(addr) ) ) {
                perror( "bind" );
                exit( 1 );
        }

	/* listen for connections */
        if ( listen( s, 5 ) == -1 ) {
                perror( "listen" );
                exit( 1 );
        }

        if ( debug )
		printf( "web500gw listening on port %d\n", port );

	return( s );
}

static wait4child()
{
#if defined(SunOS) && SunOS < 40
        union wait status;
#else
        int     status;
#endif

        if ( debug ) printf( "parent: catching child status\n" );
        while ( wait3( &status, WNOHANG | WUNTRACED, 0 ) > 0 )
                ;       /* NULL */

#ifdef __hpux
	(void) signal( SIGCHLD, wait4child );
#endif
}

static do_queries( s )
int	s;
{
	char		buf[1024], *query, *tail, *cp;
	int		len;
	FILE		*fp;
	int		rc, tblsize;
	struct timeval	timeout;
	fd_set		readfds;
	LDAP		*ld;

	if ( (fp = fdopen( s, "a+")) == NULL ) {
		perror( "fdopen" );
		exit( 1 );
	}

	tblsize = getdtablesize();
	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	FD_ZERO( &readfds );
	FD_SET( fileno( fp ), &readfds );

	if ( (rc = select( tblsize, &readfds, 0, 0, &timeout )) <= 0 )
		exit( 1 );
	/* process the request (GET, HEAD) */
	if ( fgets( buf, sizeof(buf), fp ) == NULL )
		exit( 1 );

	len = strlen( buf );
	if ( debug ) {
		fprintf( stderr, "got %d bytes\n", len );
		lber_bprint( buf, len );
	}

	/* strip of \r \n */
	if ( buf[len - 1] == '\n' ) buf[--len] = '\0';
	if ( buf[len - 1] == '\r' ) buf[--len] = '\0';
	query = strdup(buf);

	/* strip off leading white space */
	while ( isspace( *query )) ++query;
	if ((tail = strstr(query, " HTTP")) != NULL || 
	    (tail = strstr(query, " http")) != NULL) {	/* a HTTP/1.0 request */
		http = 1;
		*tail = '\0';
		while (fgets( buf, sizeof(buf), fp ) != NULL && strlen(buf) > 2) {
			/* process the following lines from client */
			/* for now: ignore them */
			if (debug > 5) fprintf (stderr, "%s", buf);
		}
	}
#ifdef __hpux
	fflush(fp);
#else
	rewind(fp);
#endif
	if (tolower(*query) == 'g' && tolower(*(query + 1)) == 'e' && tolower(*(query + 2)) == 't') {	/* HTTP GET command */
		request = GET;
		query += 3;
	} else if (tolower(*query) == 'h' && tolower(*(query + 1)) == 'e' && tolower(*(query + 2)) == 'a' && tolower(*(query + 3)) == 'd') {  /* HTTP HEAD command */
		request = HEAD;
		query += 4;
	} else {
		do_error(fp, -2, NOT_IMPLEMENTED);
		exit( 1 );
	}
        /* strip off leading white space */
	while ( isspace( *query ) || *query == '/')  ++query;
	if ( *query == 'Y' ) {  /* modify request from form */
		do_modify( fp, ++query);
		exit (0);
	}
	hex_decode(query);
	len = strlen(query);
	if (debug)
		fprintf( stderr, "--> %s.\n", query);
	if ( *query == 'H' || *query == 'E' ) {
		switch ( *query++ ) {
		case 'H':	/* help file */
			do_help( fp );
			break;
		case 'E':	/* error explanation */
			do_error( fp, query, 1 );
			break;
		}
		/* rewind(fp); */

		exit( 0 );
		/* NOT REACHED */
	}

	if ( (ld = ldap_open( ldaphost, ldapport )) == NULL ) {
		if ( debug ) perror( "ldap_open" );
		do_error( fp, LDAP_SERVER_DOWN, SERVER_ERROR);
		exit( 1 );
	}

	ld->ld_deref = LDAP_DEREF_ALWAYS;
	if ( !searchaliases )
		ld->ld_deref = LDAP_DEREF_FINDING;
	
	if (*query == 'F') {    /* Make form for modify, need user bind  */
		do_form( ld, fp, ++query);
		exit (0);
	}

	if ( (rc = ldap_simple_bind_s( ld, WEB500DN, NULL )) != LDAP_SUCCESS ) {
		if ( debug ) ldap_perror( ld, "ldap_simple_bind_s" );
		do_error( fp, rc, SERVER_ERROR);
		exit( 1 );
	}

	switch ( *query++ ) {
	case 'R':	/* Read an entry */
		do_read( ld, fp, query, 0 );
		break;
#ifdef SHOWALL
	case 'L':	/* read aLl attributes of an entry */
		do_read( ld, fp, query, 1);
		break;
#endif
	case 'S':	/* search */
		do_search( ld, fp, query );
		break;
	case 'M':	/* X.500 Menu */
		do_menu( ld, fp, query );
		break;
#ifdef JPEGTOGIF
	case 'I':	/* jpegPhoto of entry as gif Image */
		do_pict( ld, fp, query, 1);
		break;
#endif
#ifdef G3TOXBM
	case 'G':	/* G3fax photo of entry as xbm */
		do_pict( ld, fp, query, 0);
		break;
#endif
	case 'A':	/* Audio of entry */
		do_audio( ld, fp, query);
		break;
	default:
		do_menu( ld, fp, basedn );
		break;
	}

	/* rewind(fp); */
	fflush(fp);
	exit( 0 );
	/* NOT REACHED */
}

static char *pick_oc( oclist )
char	**oclist;
{
	int	i;

	if ( oclist == NULL )
		return( "unknown" );

	for ( i = 0; oclist[i] != NULL; i++ ) {
		if ( strcasecmp( oclist[i], "top" ) != 0 &&
		    strcasecmp( oclist[i], "quipuObject" ) != 0 &&
		    strcasecmp( oclist[i], "quipuNonLeafObject" ) != 0 )
			return( oclist[i] );
	}

	return( "unknown" );
}

static isnonleaf( ld, oclist, dn )
LDAP	*ld;
char	**oclist;
char	*dn;
{
	int	i, quipuobject = 0;

	if ( oclist == NULL )
		return( 0 );

	for ( i = 0; oclist[i] != NULL; i++ ) {
		if ( strcasecmp( oclist[i], "quipuObject" ) == 0 )
			quipuobject = 1;
		if ( strcasecmp( oclist[i], "quipuNonLeafObject" ) == 0 ||
		    strcasecmp( oclist[i], "externalNonLeafObject" ) == 0 )
			return( 1 );
	}

	/*
	 * not a quipu thang - no easy way to tell leaves from nonleaves
	 * except by trying to search or list, which would be expensive.
	 * for now, just guess assume all non-quipu objects are non-leaf
	 */
	return( quipuobject ? 0 : 1 );
}

static do_menu( ld, fp, dn )
LDAP	*ld;
FILE	*fp;
char	*dn;
{
	char		**s;
	char		*rdn = NULL;
        int             rc, scope, i;
        LDAPMessage     *e, *res;
        struct timeval  timeout;
        static char     *attrs[] = { "objectClass", "sn", "aliasedObjectName", 0 };
	int counter = 0;
	
	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	ld->ld_deref = LDAP_DEREF_FINDING;
	if ( (rc = ldap_search_st( ld, dn, LDAP_SCOPE_ONELEVEL,
	    "(!(objectClass=dSA))", attrs, 0, &timeout, &res )) != LDAP_SUCCESS
	    && rc != LDAP_SIZELIMIT_EXCEEDED && rc != LDAP_INSUFFICIENT_ACCESS
	    && rc != LDAP_TIMELIMIT_EXCEEDED && rc != LDAP_TIMEOUT ) {
		do_error(fp, rc, NOT_FOUND);
		return;
	}
	ld->ld_deref = LDAP_DEREF_ALWAYS;

/*	one entry: if leaf -> do_read
	it's still uneffective, in do_read the entry is searched again...

	if ( ldap_count_entries( ld, res ) == 0 ) {
		char	**oc;
		LDAPMessage	*e, *res;
		
		if ( ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*", NULL, 0, &timeout, &res ) == LDAP_SUCCESS ) {
		
			oc = ldap_get_values( ld, ldap_first_entry( ld, res ), "objectClass" );
			if (!isnonleaf(ld, oc, dn)) {
				do_read (ld, fp, dn, 0 );
				return;
			}
		}
	}	
*/
        if (http == 1) {
		PRINT_HEADER;
	}
	if (request == HEAD) {
		fflush(fp);
		exit (1);
	}
	fprintf( fp, "<HEAD><TITLE>%s</TITLE></HEAD><BODY>\n", friendly_dn(dn));

#ifdef DEFAULT_TOP_OF_BODY_FILE
	if ( (strlen(basedn) > 0) && (strcmp(dn, basedn) == 0 )) {
		char buf[4096];
		FILE *fp2;
		fp2 = fopen(DEFAULT_TOP_OF_BODY_FILE, "r" );
		if ( fp2 != NULL ) {
			while ( fgets(buf, 4095, fp2) != NULL )
				fprintf( fp, "%s\n",  buf);
			fprintf( fp, "<HR>\n");
		}
	}
#endif DEFAULT_TOP_OF_BODY_FILE

	make_header( fp, dn, 0);
	s = ldap_explode_dn( dn, 1 );
	if ( strcmp( dn, "" ) != 0 ) {	/* Not the root */
		if ( s[1] == NULL )	/* toplevel */
			rdn = ldap_friendly_name( friendlyfile, s[0], &fm );
		else
			rdn = clean_ufn( s[0]);
		fprintf( fp, "<DL><DT>Read entry of <DD><H1><A HREF=\"/R%s\">%s</A></H1></DL>\n", 
			hex_encode(dn), rdn ? rdn: s[0]);
	} else {			/* the root */
		fprintf( fp, "<H2><A HREF=\"/H\"> About the WWWeb to X.500 Gateway</A></H2><HR>\n<H1>%s</H1>\n", ROOT);
	}
	scope = make_scope(ld, dn);	/* onelevel or subtree search ? */
	fprintf( fp, "<FORM ACTION=\"/S%s\">%s search in <STRONG>%s</STRONG>: <inPUT NAME=\"%s\"> <INPUT TYPE=submit VALUE=SEARCH></FORM>\n", 
		hex_encode(dn),
		scope == LDAP_SCOPE_ONELEVEL ? "Onelevel" : "Subtree",
		rdn ? rdn : ROOT,
		scope == LDAP_SCOPE_ONELEVEL ? "O" : "S");

	ldap_value_free( s );
        if ( ldap_count_entries( ld, res ) < 1 ) {        
		if (rc == LDAP_INSUFFICIENT_ACCESS) {
			fprintf( fp, "We don't have the access right to do a global search here. ");
			fprintf( fp, "If you know who you are looking for,\r\n");
                	fprintf( fp, "try choosing the <STRONG>Search</STRONG> option listed above and\r\n" );
                	fprintf( fp, "specifying the name of the entry you want.\r\n" );
		} else if (rc == LDAP_TIMELIMIT_EXCEEDED || LDAP_TIMEOUT) {
			do_timelimit(fp);
		} else {
			fprintf( fp, "Didn't find anything below this entry! You can read the entry itself or try a search ...</BODY>\n");
		}
                return;
        }
	fflush(fp);
	fprintf( fp, "<MENU>\n" );
	counter = sort_result( ld, res, dn);
	for (i = 0 ; i < counter ; i++ ) {
		fprintf(fp,"%s",dnlist[i]->href);
		free(dnlist[i]->string);
		free(dnlist[i]->href);
	}
	fprintf(fp, "\n</MENU>\n");
	if ( ldap_result2error( ld, res, 1 ) == LDAP_SIZELIMIT_EXCEEDED ) {
		do_sizelimit(fp, 1);
	}

#ifdef DEFAULT_END_OF_BODY_FILE
	if ( strcmp(dn,basedn) == 0 ) {
		char buf[4096];
		FILE *fp2;
		fp2 = fopen(DEFAULT_END_OF_BODY_FILE, "r" );
		if ( fp2 != NULL ) {
			/* bzero(buf,4096); */
			fprintf(fp, "<HR>");
			while ( fgets(buf, 4095, fp2) != NULL ) {
				fprintf( fp, "%s\n",buf);
				bzero(buf,4096);
			}
		}
	}
#endif
	fprintf( fp, "</BODY>\r\n" );
}

static isoc( ocl, oc )
char	**ocl;
char	*oc;
{
	int	i;

	for ( i = 0; ocl[i] != NULL; i++ ) {
		if ( strcasecmp( ocl[i], oc ) == 0 )
			return( 1 );
	}

	return( 0 );
}

static int make_scope( ld, dn )
LDAP	*ld;
char	*dn;
{
	int		scope;
	char		**oc;
	LDAPMessage	*res;
	struct timeval	timeout;
	static char	*attrs[] = { "objectClass", 0 };

	if ( strcmp( dn, "" ) == 0 )
		return( LDAP_SCOPE_ONELEVEL );

	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	if ( ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*",
	    attrs, 0, &timeout, &res ) != LDAP_SUCCESS ) {
		return( -1 );
	}

	oc = ldap_get_values( ld, ldap_first_entry( ld, res ), "objectClass" );

	if ( isoc( oc, "organization" ) || isoc( oc, "organizationalUnit" ) )
		scope = LDAP_SCOPE_SUBTREE;
	else
		scope = LDAP_SCOPE_ONELEVEL;

	ldap_value_free( oc );
	ldap_msgfree( res );

	return( scope );
}

static do_search( ld, fp, query )
LDAP	*ld;
FILE	*fp;
char	*query;
{
	int		scope;
	char		*base, *filter;
	char		*filtertype;
	int		count, rc;
	struct timeval	timeout;
	LDAPFiltInfo	*fi;
	LDAPMessage	*e, *res;
	static char	*attrs[] = { "objectClass", "sn", "aliasedObjectName", 0 };
	int		counter = 0, i;
	
/* query string: base-DN?[OS]=filter 
 *     search onelevel <--||--> search subtree 
 */
	if ( (filter = strchr( query, '?' )) == NULL ) {
		explain_error( fp, "<EM>Missing filter!</EM>\n", BAD_REQUEST );
		exit( 1 );
	}
	*filter++ = '\0';
	if (*filter == '\0' || *(filter+1) != '=') {
		explain_error( fp, "<EM>Missing scope!</EM>\n", BAD_REQUEST);
		exit( 1 );
	}
	if (*filter == 'S') {
		scope = LDAP_SCOPE_SUBTREE;
	} else {
		scope = LDAP_SCOPE_ONELEVEL;
	}
	/* *filter++ = '\0'; */
	filter += 2;
	if (*filter == '\0') {
		if ( scope == LDAP_SCOPE_SUBTREE ) {
			explain_error( fp, "<EM>Missing search value!</EM> <P>Please specify a search value or a '*' to make a wildcard search.\n", BAD_REQUEST);		} else {
			explain_error( fp, "<EM>Missing search value!</EM> <P>Please specify a search value.\n", BAD_REQUEST);
		}
		exit( 1 );
	}
	base = query;

	filtertype = (scope == LDAP_SCOPE_ONELEVEL ? "web500gw onelevel" :
	    "web500gw subtree");
	ld->ld_deref = (scope == LDAP_SCOPE_ONELEVEL ? LDAP_DEREF_FINDING :
	    LDAP_DEREF_ALWAYS);
	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	
	friendlyDesc = NULL;
	for ( fi = ldap_getfirstfilter( filtd, filtertype, filter ); fi != NULL;
	    fi = ldap_getnextfilter( filtd ) ) {
		if ( (rc = ldap_search_st( ld, base, scope, fi->lfi_filter,
		    attrs, 0, &timeout, &res )) != LDAP_SUCCESS
		    && rc != LDAP_SIZELIMIT_EXCEEDED ) {
			do_error(fp, rc, NOT_FOUND);
			return( 1 );
		}
		if ( (count = ldap_count_entries( ld, res )) != 0 ) {
			friendlyDesc = fi->lfi_desc;
			break;
		}
	}
	ld->ld_deref = LDAP_DEREF_ALWAYS;

	if ( count == 0 ) {
		if (http == 1) {
			fprintf(fp, "HTTP/1.0 404 Not found\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: text/html\r\n\r\n", version);
		}
		if (request == HEAD) {
			fflush(fp);
			exit (1);
		}	
		fprintf( fp, "<HEAD><TITLE>No search results: %s</TITLE></HEAD>\n", filter);
		fprintf( fp, "<BODY><H2>Didn't find anything...</H2> for <STRONG>%s</STRONG> in <STRONG>%s</STRONG></BODY>\n", filter, strlen(base) == 0 ? ROOT : clean_ufn(ldap_dn2ufn(base)));
		return( 0 );
	}
#ifdef SHOWSINGLEENTRY
	else if ( count == 1 ) {
		e = ldap_first_entry( ld, res );
		if ( e != NULL ) {
			char *dn, **oc, **aliasdn;
			oc = ldap_get_values(ld, e, "objectClass");
			aliasdn = ldap_get_values(ld, e, "aliasedObjectName");
#ifdef DEREFALIAS
			if ( aliasdn && *aliasdn) 
				deref(ld, dn, oc);
			else
#endif
				dn = ldap_get_dn(ld, e);
			if ( dn ) {
				if ( isnonleaf(ld, oc, dn) )
					do_menu(ld, fp, dn);
				else
					do_read(ld, fp, dn, 0);
				return (0);
			}
		}
	}
#endif /* SHOWSINGLEENTRY */

	if (http == 1)
		PRINT_HEADER;
	if (request == HEAD) {
		fflush(fp);
		exit (1); 
        }

	fprintf( fp, "<HEAD><TITLE>Search results: %s</TITLE></HEAD>\n", filter );
        fprintf( fp, "<BODY>Search results for <STRONG>\"%s\"</STRONG> in <STRONG>\"%s\"</STRONG> (%d %s by <STRONG>%s</STRONG>)<MENU>\r\n", 
		filter, strlen(base) == 0 ? ROOT : friendly_dn(base),
		count, count == 1 ? "entry" : "entries", friendlyDesc);

	counter = sort_result( ld, res, base );
	for (i = 0; i < counter ; i++ ) {
		fprintf(fp,"%s",dnlist[i]->href);
		free(dnlist[i]->string);
		free(dnlist[i]->href);
	}
	fprintf( fp, "</MENU>\n");

	if ( ldap_result2error( ld, res, 1 ) == LDAP_SIZELIMIT_EXCEEDED ) {
		do_sizelimit(fp, 0);
	}
	fprintf( fp, "</BODY>\r\n" );
}

static print_attr( ld, fp, label, printattr, e, flag)
LDAP		*ld;
FILE		*fp;
char		*label;
char		*printattr;
LDAPMessage	*e;
int		flag;
{
	char	**val, *domain, *user, *cp;
	int	i, gotone = 0, firstline = 1;
	static char *nlabel;
	nlabel = ldap_friendly_name(friendlyfile, label, &fm); 

	if ( (val = ldap_get_values( ld, e, printattr )) == NULL )
		return;
	
	fprintf( fp, "\n<DT><B>%s</B><DD>", nlabel );
	for ( i = 0; val[i] != NULL; i++ ) {
#ifdef URL_SUPPORT
		if (flag == URL) {
			if ((cp = strchr(val[i], '$')) != NULL) {
				*cp++ = '\0';
				fprintf(fp, "<A HREF=\"%s\"> %s</A><BR>\n", cp, val[i]);
			} else 
				fprintf( fp, "%s<BR>\n", val[i]);
		} else
#endif /* URL_SUPPORT */
		if ( flag == MULTILINE ) {
			char	*s, *p;

			if ( gotone )
				fprintf( fp, "<DT><B>%s</B><DD>", nlabel);
			p = s = val[i];
			while ( s = strchr( s, '$' ) ) {
				*s++ = '\0';
				while ( isspace( *s ) )
					s++;
				if ( firstline ) {
					fprintf( fp, "%s<BR>\n", p );
					firstline = 0;
				} else {
					fprintf( fp, "%s<BR>\n", p );
				}
				p = s;
			}
			if ( firstline ) {
				fprintf( fp, "%s<BR>\n", p );
			} else {
				fprintf( fp, "%s", p );
			}
			gotone = 1;
			firstline = 1;
		} else if (flag == DATE) {
#ifdef RUTGERS
			if ( strcmp(printattr, "lastUpdateTime") == 0) 
				fprintf( fp, "%s<BR>\n", format_date(&val[i], "%r on %A, %B %d %Y"));
			else
#endif /* RUTGERS */
				fprintf( fp, "%s<BR>\n", format_date(&val[i], "%A, %d-%h-%y %T GMT"));

#ifdef MAIL_FINGER
		} else if (flag == FINGER) {
#ifdef FINGER_CMD
			fprintf( fp, "<A HREF=\"%s%s\">%s</A><BR>\n",
				FINGER_CMD, val[i], val[i]);
#else
			/* default gopher <-> finger thing */
			user = strdup(val[i]);
			if ((domain = strchr(user, '@')) != NULL) {
				*domain++ = '\0';
				fprintf( fp, "<A HREF=\"gopher://%s:79/0%s\">%s</A><BR>\n",				domain, user, val[i]);
			} else
				fprintf( fp, "%s<BR>\n", val[i]);
#endif /* FINGER_CMD */
#endif /* MAIL_FINGER */
		} else if (flag == HREF) {
			char    *ufn;
			ufn = friendly_dn( val[i] );
			fprintf( fp, "<A HREF=\"/R%s\">%s</A><BR>\n",
				hex_encode(val[i]), ufn ? ufn : val[i]);
			if ( ufn ) {
				free( ufn );
			}
		} else if ( firstline ) {
			fprintf( fp, "%s", val[i] );
			firstline = 0;
		} else {
			fprintf( fp, "<BR>\n%s", val[i] );
		}
	}
	ldap_value_free( val );
}

static form_attr( ld, fp, label, formattr, e, multiline, add_empty )
LDAP            *ld;
FILE            *fp;
char            *label;
char            *formattr;
LDAPMessage     *e;
int             multiline;
int             add_empty;
{
        char    **val, *s;
        char    buffer[1024];
        int     i, gotone = 0, line = 0;
        static char *nlabel;
        nlabel = ldap_friendly_name(friendlyfile, label, &fm);
        if ( (val = ldap_get_values( ld, e, formattr )) == NULL ) {
		if (add_empty > 0)
			if(multiline)
				fprintf( fp, "<DT><B>%s</B><DD><INPUT NAME=\"%s=\" SIZE=30,4  ><BR>\n", nlabel, label);
			else
                		fprintf( fp, "<DT><B>%s</B><DD><INPUT NAME=\"%s=\" SIZE=30 ><BR>\n", nlabel, label);
                return;
	}
        fprintf( fp, "\n<DT><B>%s</B><DD>", nlabel );
	buffer[0] ='\0';
        for ( i = 0; val[i] != NULL; i++ ) {
                if ( multiline ) {
                        char    *s, *p;
                	buffer[0] ='\0';
                        if ( gotone )
                                fprintf( fp, "<DT><B>%s</B><DD>", nlabel);
                        p = s = val[i];
                        while ( s = strchr( s, '$' ) ) {
                                *s++ = '\0';
                                while ( isspace( *s ) )
                                        s++;
                                if ( line == 0 ) {
                                        sprintf(buffer, "%s", p);
                                        line++;
                                } else {
                                        sprintf(buffer, "%s\n%s", buffer, p);
					line++;
                                }
                                p = s;
                        }
			if (line++ == 0) {
				sprintf(buffer, "%s", p);
			} else {
				sprintf(buffer, "%s\n%s", buffer, p);
			}
			s = form_encode(buffer);
                        fprintf( fp, "<INPUT NAME=\"%s=%s\" SIZE=30,%d VALUE=\"%s\"> <BR>\n", label, s, line + 1, s);
                        gotone = 1;
                        line = 0;
                } else {
			s = form_encode(val[i]);
                        fprintf( fp, "<INPUT NAME=\"%s=%s\" SIZE=\"%d\" VALUE=\"%s\"><BR>\n", label, s, strlen(val[i]) > 30 ? strlen(val[i]) + 3 : 30, s);
                }
        }
	if (add_empty > i)
		fprintf( fp, "<INPUT NAME=\"%s=\" SIZE=30 ><BR>\n", label);
        ldap_value_free( val );
}

static do_read( ld, fp, dn, all )
LDAP	*ld;
FILE	*fp;
char	*dn;
int	all;
{
	int		rc, photo = 0, i = 0, nonleaf;
	char    	**val, **s, *oc;
	char		*a, *rdn, label[256], *date;
	time_t		t;
	LDAPMessage	*res, *e;
	struct timeval	timeout;
	BerElement	*ber;
	struct tm	tm, *ntm;

	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	if ( (rc = ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*",
	    NULL, 0, &timeout, &res )) != LDAP_SUCCESS ) {
		do_error(fp, rc, NOT_FOUND);
		return;
	}
	if ( (e = ldap_first_entry( ld, res )) == NULL ) {
		do_error(fp, -2, SERVER_ERROR);
		return;
	}
	if (http == 1) {
		fprintf(fp, "HTTP/1.0 %d OK\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: text/html\r\n", DOCUMENT_FOLLOWS, version);
		if ((val = ldap_get_values( ld, e, "lastModifiedTime")) != NULL) {
			date = format_date(val, "%A, %d-%h-%y %T GMT");
			if (debug) fprintf(stderr, "Last-Modified: %s -> %s\n", val[0], date);
			fprintf(fp, "Last-Modified: %s\r\n", date);
		}
		fprintf(fp, "\r\n");
	}
        if (request == HEAD) {                
		fflush(fp);
		exit (1);
        }
	dn = ldap_get_dn( ld, e );

        if ( strcmp( dn, "" ) != 0 ) {  /* Not the root */
		s = ldap_explode_dn( dn, 1 );
                if ( s[1] == NULL )   /* toplevel */
			rdn = ldap_friendly_name( friendlyfile, s[0], &fm );
		else 
			rdn = clean_ufn(s[0]);
	} else
		rdn = ROOT;

	fprintf( fp, "<HEAD><TITLE>%s</TITLE></HEAD>\r\n<BODY>\r\n", friendly_dn( dn ) );
	make_header( fp, dn, 0 );
	if (friendlyDesc != NULL)
		fprintf(fp, "Found match on <STRONG>%s</STRONG>\n", friendlyDesc);
	fprintf( fp, "<DL>");
	val = ldap_get_values( ld, e, "objectClass" );
	if ( val != NULL && (nonleaf = isnonleaf(ld, val, dn )) )
		fprintf( fp, "<DT>Move to <DD><H1><A HREF=\"/M%s\">%s</A></H1>\r\n",
			hex_encode(dn), rdn);
	else
		fprintf( fp, "<H1>%s</H1>\r\n", rdn);
#ifdef JPEGTOGIF
	/* look for a photo, prefer jpeg */
	if ((val = ldap_get_values( ld, e, "jpegPhoto" )) != NULL ) {
		fprintf( fp, "<DD><IMG ALT=\"JPEG photo\" SRC=\"/I%s\">\n", hex_encode(dn));
		photo = 1;
	}
#endif
#ifdef G3TOXBM
	if ((val = ldap_get_values( ld, e, "photo" )) != NULL )
		if (photo)
			fprintf(fp, "<A HREF=\"/G%s\">Another Photo (G3 Fax)</A>\r\n", hex_encode(dn));
		else
			fprintf(fp, "<DD><IMG ALT=\"G3 Fax photo\" SRC=\"/G%s\">\n", hex_encode(dn));
#endif
	for ( a = ldap_first_attribute( ld, e, &ber ); a != NULL;
	    a = ldap_next_attribute( ld, e, ber ) ) {
		sprintf( label, "%s", a );
#ifdef SHOWALL		
		if (!all && ( strcasecmp( a, "acl" ) == 0 ||
#else
		if ((strcasecmp( a, "acl" ) == 0 ||
#endif
		    strcasecmp( a, "xacl" ) == 0 ||
		    strncasecmp( a, "lastModified", 12 ) == 0 ||
		    strcasecmp( a, "eDBInfo" ) == 0 ||
		    strcasecmp( a, "iattr" ) == 0 ||
		    strcasecmp( a, "sacl" ) == 0 ||
		    strcasecmp( a, "lacl" ) == 0 ||
		    strcasecmp( a, "masterDSA" ) == 0 ||
		    strcasecmp( a, "slaveDSA" ) == 0 ||
		    strcasecmp( a, "objectClass" ) == 0 ||
		    strcasecmp( a, "userPassword" ) == 0 ||
		    strcasecmp( a, "treeStructure" ) == 0 ||
		    strcasecmp( a, "sn" ) == 0 ||
		    strcasecmp( a, "krbName" ) == 0 ||
		    strcasecmp( a, "delegate" ) == 0 ||
#ifdef RUTGERS
		    strcasecmp( a, "source" ) == 0 ||
#endif
		    isdigit( a[0] )) ) {
			continue;
		} else if (strcasecmp( a, "photo" ) == 0 ||
			   strcasecmp( a, "jpegPhoto" ) == 0 ) {
			/* allready shown */
			continue;
		} else if (strcasecmp( a, "audio" ) == 0) {
		/* internal-gopher-sound is Mosaic hack */
			fprintf(fp, "</DL><A HREF=\"/A%s\"> <IMG ALT=\"AUDIO\" SRC=\"internal-gopher-sound\"> %s</A><DL>\n", hex_encode(dn), ldap_friendly_name(friendlyfile, "audio", &fm));
#ifdef MAIL_FINGER
		} else if (strcasecmp( a, "mail" ) == 0) {
			print_attr( ld, fp, label, a, e, FINGER);
#endif
#ifdef URL_SUPPORT
		} else if (strcasecmp( a, "labeledURL" ) == 0 ) {
			print_attr( ld, fp, label, a, e, URL);
#endif
		} else if ( strcasecmp( a, "postalAddress" ) == 0 ||
#ifdef UOFM
		    strcasecmp( a, "multiLineDescription" ) == 0 ||
#endif
		    strcasecmp( a, "homePostalAddress" ) == 0 ) {
			print_attr( ld, fp, label, a, e, MULTILINE);
		} else if ( strcasecmp( a, "seeAlso" ) == 0 ||
			    strcasecmp( a, "member" ) == 0 ||
			    strcasecmp( a, "owner" ) == 0 ||
			    strcasecmp( a, "roleOccupant" ) == 0 ||
			    strcasecmp( a, "manager" ) == 0 ||
			    strcasecmp( a, "secretary" ) == 0 ||
			    strcasecmp( a, "associatedName" ) == 0 ||
			    strcasecmp( a, "lastModifiedBy" ) == 0 ||
			    strcasecmp( a, "masterDSA" ) == 0 ||
			    strcasecmp( a, "slaveDSA" ) == 0 ||
#ifdef UOFM		
			    strcasecmp( a, "memberOfGroup" ) == 0 ||
			    strcasecmp( a, "errorsTo" ) == 0 ||
			    strcasecmp( a, "requestsTo" ) == 0 ||
#endif /* UOFM */
			    strcasecmp( a, "responsiblePerson" ) == 0) {
			print_attr( ld, fp, label, a, e, HREF);
		} else if (strcasecmp( a, "lastModifiedTime") == 0
#ifdef RUTGERS
			|| strcasecmp(a, "lastUpdateTime") == 0
#endif
								) {
			/* write readable date */
			print_attr( ld, fp, label, a, e, DATE);
		} else {
			print_attr( ld, fp, label, a, e, DEFAULT);
		}
	}
#ifdef SHOWALL
	if (!all) {
		fprintf( fp, "</DL><A HREF=\"/L%s\"><B>See all attributes</B></A>\n",
		hex_encode(dn));
	}
#endif
	fprintf( fp, "</DL>\n");
#ifdef MODIFY
	if ((oc = pick_oc( ldap_get_values( ld, e, "objectClass" ) )) != NULL &&
		strcasecmp(dn + strlen(dn) - strlen(basedn), basedn) == 0) {
		for (i = 0; mod_oc[i]; i++) {
			if (strcasecmp(oc, mod_oc[i]) == 0) {
	 			fprintf( fp, "<HR><TT>Beware, this is still <B>EXPERIMENTAL</B>!!!</TT><P><FORM ACTION=\"F%s\">\n\
	 		Enter the password of this entry\
		 	<INPUT TYPE=\"password\" NAME=\"userPassword\"><BR>\
		 	if you want to  <INPUT TYPE=\"submit\" VALUE=\"MODIFY\"> it.\
		 	</FORM>\n", hex_encode(dn));
				break;
		    	}
		}
	}
#endif MODIFY
	fprintf( fp, "</BODY>\n");
}

static do_form( ld, fp, query )
LDAP    *ld;
FILE    *fp;
char    *query;
{
        int             rc, i, photo = 0;
        char            *dn = query, *pw;
        char            *ufn;
        char            **val;
        char            *a, *s, label[256];
        LDAPMessage     *res, *e;
        struct timeval  timeout;
        BerElement      *ber;


        if ( (pw = strchr( dn, '?' )) == NULL ) {
                fprintf( fp, "Missing Password!\r\n" );
                exit( 1 );
        }
        *pw++ = '\0';
        if (strncmp(pw, "userPassword=", 13) == 0)
                pw += 13;
        else {
                fprintf( fp, "Unknown argument %s!\r\n", pw);
                exit ( 1 );
        }
	if (strlen(pw) == 0) {
		/* we need a password for simple auth */
		do_error( fp, LDAP_INVALID_CREDENTIALS, FORBIDDEN);
		exit( 1 );
	}
        if ( (rc = ldap_simple_bind_s( ld, dn, pw )) != LDAP_SUCCESS ) {
                if ( debug ) ldap_perror( ld, "ldap_simple_bind_s" );
                do_error( fp, rc, FORBIDDEN);
                exit( 1 );
        }
        if (debug) fprintf(stderr, "BOUND as %s\n", dn);
        timeout.tv_sec = TIMEOUT;
        timeout.tv_usec = 0;
        if ( (rc = ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*",
            0, 0, &timeout, &res )) != LDAP_SUCCESS ) {
                do_error(fp, rc, NOT_FOUND);
                return;
        }
        if ( (e = ldap_first_entry( ld, res )) == NULL ) {
                do_error(fp, -2, SERVER_ERROR);
                return;
        }
        dn = ldap_get_dn( ld, e );
        ufn = friendly_dn( dn );
        if (http == 1) {
		PRINT_HEADER;
        }
        fprintf( fp, "<HEAD><TITLE>Modify %s</TITLE></HEAD>\r\n<BODY>\r\n", ufn
);
        fprintf( fp, "<FORM ACTION=\"Y%s\">\n", hex_encode(dn));
	fprintf( fp, "<INPUT TYPE= \"radio\" NAME=\"oldPassword\" VALUE=\"%s\" CHECKED><TT>Beware, this is still EXPERIMENTAL!!!</TT>\n",
		hex_encode(pw));
        fprintf( fp, "<H1>%s</H1><DL>\r\n", ufn );
        free( ufn );
#ifndef OLD
	for ( a = mod_attr[0].attr_name, i = 0; a = mod_attr[i].attr_name; i++ ) {
                if ( strcasecmp( a, "homePostalAddress" ) == 0
#ifdef UOFM
                    || strcasecmp( a, "multiLineDescription" ) == 0
#endif
                    || strcasecmp( a, "postalAddress" ) == 0) 
			if (mod_attr[i].attr_numb == 0)
                        	print_attr( ld, fp, a, a, e, MULTILINE );
			else
                        	form_attr( ld, fp, a, a, e, 1, mod_attr[i].attr_numb  );
                else if (mod_attr[i].attr_numb == 0)
                	print_attr( ld, fp, a, a, e, DEFAULT);
		else
                	form_attr( ld, fp, a, a, e, 0, mod_attr[i].attr_numb );
	}
#else
        for ( a = ldap_first_attribute( ld, e, &ber ); a != NULL;
            a = ldap_next_attribute( ld, e, ber ) ) {
                sprintf( label, "%s", a );

                if ( strcasecmp( a, "acl" ) == 0 ||
                    strcasecmp( a, "xacl" ) == 0 ||
                    strncasecmp( a, "lastModified", 12 ) == 0 ||
                    strcasecmp( a, "eDBInfo" ) == 0 ||
                    strcasecmp( a, "iattr" ) == 0 ||
                    strcasecmp( a, "sacl" ) == 0 ||
                    strcasecmp( a, "lacl" ) == 0 ||
                    strcasecmp( a, "audio" ) == 0 ||
                    strcasecmp( a, "masterDSA" ) == 0 ||
                    strcasecmp( a, "slaveDSA" ) == 0 ||
                    strcasecmp( a, "objectClass" ) == 0 ||
                    strcasecmp( a, "userPassword" ) == 0 ||
                    strcasecmp( a, "treeStructure" ) == 0 ||
                    strcasecmp( a, "photo" ) == 0 ||
                    strcasecmp( a, "jpegPhoto" ) == 0 ||
		    strcasecmp( a, "postalAddress") == 0 ||
		    strcasecmp( a, "seeAlso") == 0 ||
		    strcasecmp( a, "sn" ) == 0 ||
                    isdigit( a[0] ) ) {
                        continue;
                } else if ( strcasecmp( a, "homePostalAddress" ) == 0
#ifdef UOFM
                    || strcasecmp( a, "multiLineDescription" ) == 0 
#endif
                    ) {
                        form_attr( ld, fp, label, a, e, 1, 0 );
                } else if (strcasecmp( a, "cn") == 0) {
			print_attr( ld, fp, label, a, e, DEFAULT );
		} else if (strcasecmp( a, "description") == 0 ||
			   strcasecmp( a, "info") == 0) {
			form_attr( ld, fp, label, a, e, 0, 3);
		} else {
                        form_attr( ld, fp, label, a, e, 0, 0 );
                }
        }
#endif
        fprintf( fp, "</DL><InPut TYPE=\"reset\" VALUE=\"RESET Values\"> <InPut TYPE=\"submit\" VALUE=\"MODIFY\"></FORM></BODY>\n");
	fflush(fp);
}

static do_modify( fp, query )
FILE    *fp;
char    *query;
{
        char *dn = query, *ufn, *pw, *what, *next, *val, *oldval, *cp;
        char b, c;
        int     rc, changes = 0, change = 0;
        LDAP    *ld;
	static char    *value[2];
	static LDAPMod mod;
	static LDAPMod *mods[2] = { &mod };
        LDAPMessage     *res, *e;
	struct timeval  timeout;


	/* query: DN?oldPassword=oldpw&att1=oldval1=val1&att2=oldval2=val2&...
	 * or:    DN?oldPassword=oldpw&att1%3Doldval1=val1&att2%3Doldval2=... 
	 */

	/* rewind(fp); */
        if ( (what = strchr( dn, '?' )) == NULL ) {
                explain_error( fp, "<EM>Missing DN!</EM>", BAD_REQUEST );
                exit( 1 );
        }
        *what++ = '\0';
        hex_decode(dn);
        if (debug) fprintf(stderr, "What: %s\n", what);	
	/* At first there should be the old userPassword */
	if ( (pw = strstr( what, "oldPassword")) == NULL ) {
		explain_error( fp, "<EM>Missing password!</EM>", BAD_REQUEST);
		exit ( 1 );
	}
        pw += 12;		/* strlen("oldPassword") + 1 */
	/* skip to the first real attribute */
	if ( (what = strchr(pw, '&'))  == NULL ) {
		explain_error( fp, "<EM>Unknown query!</EM>", BAD_REQUEST);
		exit( 1 );
	}
	*what++ = '\0';
	hex_qdecode(pw);
        if (debug) fprintf(stderr, "\ndo_modify: DN = %s PW = #######  CONTENT =\n%s\n ", dn, pw, what );
        if ( (ld = ldap_open( ldaphost, ldapport )) == NULL ) {
                if ( debug ) perror( "ldap_open" );
                do_error( fp, LDAP_SERVER_DOWN, SERVER_ERROR);
                exit( 1 );
        }
        if ( (rc = ldap_simple_bind_s( ld, dn, pw )) != LDAP_SUCCESS ) {
                if ( debug ) ldap_perror( ld, "ldap_simple_bind_s" );
                do_error( fp, rc, FORBIDDEN);
                exit( 1 );
        }
        if (debug) fprintf(stderr, "BOUND as %s.\n", dn);
        if (http == 1) {
		PRINT_HEADER;
        }
	if (request == HEAD) {
		fflush(fp);
		exit (1);
	}
	ufn = friendly_dn( dn );
	fprintf(fp, "<TITLE>Modify results: %s</TITLE><BODY>\
		<H2>Modify results for %s</H2>Changes:<DL>\n", ufn, ufn);
	free(ufn);

        while (what) {
                if ((next = strchr(what, '&')) != NULL) {
			*next++ = '\0';
		} else {	/* last in query */
			next = NULL;
		}
		if ((val = strrchr(what, '=')) == NULL) {
			/* new value */
			fprintf( fp, "<P>Error in query: <EM>No value for %s!</EM><P>", hex_qdecode(what));
			exit (1);
		}
		*val++ = '\0';
		hex_qdecode(what);
		hex_qdecode(val);
		if ((oldval = strchr(what, '=')) == NULL) {
			/* old value */
			fprintf( fp, "<P>Error in query: <EM>No oldvalue for %s!</EM><P>\n", what);
			exit (1);
		}
		*oldval++ = '\0';
		if (strcmp(oldval, val) == 0 ) {	/* no changes */
			what = next;
			continue;
		}
		if (strcasecmp(what, "homePostalAddress") == 0) {
			/* multiline */
			cp = oldval;
			while ((cp = strchr(cp, '\n')) != NULL) *cp = '$';
			cp = val;
			while ((cp = strchr(cp, '\n')) != NULL) *cp = '$';
                }
		if (debug)
			fprintf(stderr, "what = %s, oldval = %s, val = %s\n", what, oldval, val);

		/* there is something to do:
		 * - delete the old value
		 * - add the new value if not empty */
		mod.mod_type = what;
		value[1] = NULL;
		mod.mod_values = value;
		if (strlen(oldval) > 0) {
			mod.mod_op = LDAP_MOD_DELETE;
			value[0] = oldval;
			if ((rc = ldap_modify_s(ld, dn, mods)) != LDAP_SUCCESS) {
				if ( debug ) ldap_perror( ld, "ldap_modify_s: DELETE");
			        fprintf( fp, "<B>Error during modify operation:</B><P>\nDELETE old value <TT>%s</TT> from attribute <TT>%s</TT><P>The error code was <EM> %d: %s.</EM><P>", oldval, what, rc, ldap_err2string( rc ) );

				exit( 1 );
			}
		}
		if (strlen(val) > 0) {
			/* now add the new value */
			mod.mod_op = LDAP_MOD_ADD;
			value[0] = val;
			if (debug) fprintf(stderr, "trying: %s = %s.\n", what, val);
			if ((rc = ldap_modify_s(ld, dn, mods)) != LDAP_SUCCESS) {
				if ( debug ) ldap_perror( ld, "ldap_modify_s: ADD");
				fprintf( fp, "<B>Error during modify operation:</B><P>\nADD new value <TT>%s</TT> to attribute <TT>%s</TT><P>The error code was <EM> %d: %s.</EM><P>", val, what, rc, ldap_err2string( rc ) );
				if (strlen(oldval) > 0 && rc != LDAP_INSUFFICIENT_ACCESS) {
					/* try to reset to old value */
					value[0] = oldval;
					if ((rc = ldap_modify_s(ld, dn, mods)) != LDAP_SUCCESS) {
						fprintf( fp, "<B>Fatal Error:</B><BR>Couldn't reset the old value <TT>%s</TT> to attribute <TT>%s</TT>!<P>\n",
						oldval, what, rc, ldap_err2string(rc));

						exit( 1 );
					} else {
						fprintf( fp, "Reset the old  value <TT>%s</TT> to attribute <TT>%s</TT><P>\n", oldval, what);
					}
				}
				what = next;
				continue;
			}
		}
		if (debug) fprintf(stderr, "MOD: %s = %s.\n", what, val);
		changes++;
		fprintf(fp, "<DT><B>%s</B> <DD>%s <TT>(%s)</TT>\n",
			ldap_friendly_name(friendlyfile, what, &fm), value[0],
			mod.mod_op == LDAP_MOD_DELETE ? "deleted" : strlen(oldval) > 0 ? "replaced" : "added");
		what = next;
        }
	fprintf(fp, "</DL>%d change%s%s!\n", changes, 
		changes != 1 ? "s" : "", changes > 0 ? " successfully done" : "");
	if (changes > 0) {
		fprintf(fp, "<UL><LI><B><A HREF=\"/R%s\">Read this modified entry</A>\
<LI><A HREF=\"/F%s?userPassword=%s\">Modify this entry again</A></B></UL>\n",
		 hex_encode(dn), hex_encode(dn), pw);
	}
	 fprintf(fp, "</BODY>");
	 fflush(fp);
}

static do_pict( ld, fp, dn, type )
LDAP	*ld;
FILE	*fp;
char	*dn;
int	type;
{
	int		rc, i;
	struct berval	**val;
	char	cmd[128], buffer[1024];
	char		*a, *s, label[256];
	FILE		*op, *tp;
	LDAPMessage	*res, *e;
	struct timeval	timeout;
	BerElement	*ber;
	struct stat	st;

	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	if ( (rc = ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*",
	    NULL, 0, &timeout, &res )) != LDAP_SUCCESS ) {
		do_error(fp, rc, NOT_FOUND);
		return;
	}

	if ( (e = ldap_first_entry( ld, res )) == NULL ) {
		do_error(fp, -2, SERVER_ERROR);
		return;
	}
	if ( (val = ldap_get_values_len( ld, e, type == 0 ? "photo" : "jpegPhoto"  )) == NULL )
		return;
	tp = fopen(s=tempnam("/tmp", "web500"), "w+");
	if (type == 0) {  /* g3fax photo -> xbm */
		sprintf(cmd, "%s > %s", G3TOXBM, s);
		if (debug) fprintf(stderr, "%d bytes FAX!! %s\n", val[0]->bv_len, 
				cmd);
	} else {	/* jpeg -> gif */
		sprintf(cmd, "%s > %s", JPEGTOGIF, s);
		if (debug) fprintf(stderr, "%d bytes JPEG!! %s\n", val[0]->bv_len,
				cmd);
	}
	if (http == 1) {
		fprintf(fp, "HTTP/1.0 %d OK\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: image/%s\r\n", DOCUMENT_FOLLOWS, version, type == 1 ? "gif" : "x-xbitmap", val[0]->bv_len);
	}
        if (request == HEAD) {                
		fflush(fp);
		exit (1);                                   
        }                          
	if ((op = popen(cmd, "w")) == NULL ) 
		 return;
	fwrite(val[0]->bv_val, val[0]->bv_len, 1, op);
	pclose(op);
	if (stat(s, &st) == 0 && http == 1) {
		fprintf(fp, "Content-length: %d\r\n", st.st_size);
		if (debug) fprintf(stderr, "Image size: %d\n", st.st_size);
	}
	fprintf(fp, "\r\n");

	while (i = fread(buffer, 1, 1024, tp)) fwrite(buffer, 1, i, fp);
	fclose(tp);
	if (unlink(s) == -1) {
		if (debug) perror("Couldn't unlink temp image file");
	}
	fflush(fp);
}

static do_audio( ld, fp, dn, type )
LDAP	*ld;
FILE	*fp;
char	*dn;
int	type;
{
	int		rc;
	struct berval	**val;
	LDAPMessage	*res, *e;
	struct timeval	timeout;
	BerElement	*ber;

	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	if ( (rc = ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectClass=*",
	    NULL, 0, &timeout, &res )) != LDAP_SUCCESS ) {
		do_error(fp, rc, NOT_FOUND);
		return;
	}

	if ( (e = ldap_first_entry( ld, res )) == NULL ) {
		do_error(fp, -2, SERVER_ERROR);
		return;
	}
	if ( (val = ldap_get_values_len( ld, e, "audio" )) == NULL )
		return;
	if (http == 1) {
		fprintf(fp, "HTTP/1.0 %d OK\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: audio/basic\r\nContent-length: %d\r\n\r\n", DOCUMENT_FOLLOWS, version, val[0]->bv_len);
	}
        if (request == HEAD) {                
		fflush(fp);
		exit (1);                                   
        }                          
	fwrite(val[0]->bv_val, val[0]->bv_len, 1, fp);
	fflush(fp);
}

static do_help( fp )
FILE	*fp;
{
	FILE	*op;
	char	line[BUFSIZ];
	int	sent = 0;

        if (http == 1) {
                PRINT_HEADER;
        }
        if (request == HEAD) {
                fflush(fp);
                exit (1);
        }

	if ( (op = fopen( helpfile, "r" )) == NULL ) {
		fprintf( fp, "Cannot access helpfile (%s)\r\n", helpfile );
		return;
	}

	while ( fgets( line, sizeof(line), op ) != NULL ) {
		sent++;
		fprintf( fp, "%s", line );
	}
	fclose( op );
	fflush( fp );
	if (debug) fprintf( stderr, "help done: %d lines\n", sent);
}

static do_sizelimit( fp, type )
FILE	*fp;
int	type;
{
	if ( type == 0 ) {
		fprintf( fp, "The query you specified was not specific enough, causing a size limit\r\n" );
		fprintf( fp, "to be exceeded and the first several matches found to be returned.\r\n" );
		fprintf( fp, "If you did not find the match you were looking for, try issuing a more\r\n" );
		fprintf( fp, "specific query, for example one that contains both first and last name.\r\n" );
	} else {
		fprintf( fp, "Not all entries could be returned because a size limit was exceeded.\r\n" );
		fprintf( fp, "There is no way to defeat this feature, but if you know who you are\r\n" );
		fprintf( fp, "looking for, try choosing the <STRONG>Search</STRONG> option listed above and\r\n" );
		fprintf( fp, "specifying the name of the entry you want.\r\n" );
	}
}

static do_timelimit( fp )
FILE	*fp;
{
	fprintf( fp, "A timeout occured during searching X.500\r\n");
}

static do_error( fp, code, status )
FILE	*fp;
int	code;
int	status;
{
	char *s;
       	if (http == 1) {
		switch (status) {
		case BAD_REQUEST:	s = "Bad request"; break;
		case AUTH_REQUIRED:	s = "Authorization required"; break;
		case FORBIDDEN:		s = "Forbidden"; break;
		case NOT_FOUND:		s = "Not found"; break;
		case SERVER_ERROR:	s = "Server error"; break;
		case NOT_IMPLEMENTED:	s = "Not implemented"; break;
		default:		s = "Unknown error";
		}
		fprintf(fp, "HTTP/1.0 %d %s\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: text/html\r\n\r\n", status, s, version);
        }
        if (request == HEAD) {
		fflush(fp);
               	exit (1);
        }
	fprintf( fp, "<HEAD><TITLE>Error</TITLE></HEAD><BODY>\n");
	fprintf( fp, "<H2>X.500 - WWW Error: %s</H2>\n", s);
	fprintf( fp, "An error occurred searching X.500 <P>  The error code was");
	fprintf( fp, " <EM> %d: %s.</EM><P>", code, ldap_err2string( code ) );
	fprintf( fp, "No additional information is available.\r\n" );
	fprintf( fp, "</BODY>\r\n" );
	fflush( fp );
}

static explain_error ( fp, error, status )
FILE    *fp;
char	*error;
int	status;
{
        char *s;
        if (http == 1) {    
                switch (status) {
                case BAD_REQUEST:       s = "Bad request"; break;
                case AUTH_REQUIRED:     s = "Authorization required"; break;
                case FORBIDDEN:         s = "Forbidden"; break;   
                case NOT_FOUND:         s = "Not found"; break;
                case SERVER_ERROR:      s = "Server error"; break;              
                case NOT_IMPLEMENTED:   s = "Not implemented"; break;
                default:                s = "Unknown error";                  
                }                                                    
                fprintf(fp, "HTTP/1.0 %d %s\r\nServer: %s\r\nMIME-Version: 1.0\r\nContent-type: text/html\r\n\r\n", status, s, version);
        }                                           
        if (request == HEAD) {                   
		fflush(fp);
                exit (1);               
        }
        fprintf( fp, "<HEAD><TITLE>Error</TITLE></HEAD><BODY>\n");
	fprintf( fp, "<H2>Error: %s</H2>\n", s);
	fprintf( fp, "%s</BODY>\n", error);
}	


/*
 *  Utilities for dealing with HTML junk
 */

static int acceptable[256];
static int acceptable_inited = 0;

void init_acceptable()
{
    unsigned int i;
    char * good =
      "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./-_$";
    for(i=0; i<256; i++) acceptable[i] = 0;
    for(;*good; good++) acceptable[(unsigned int)*good] = 1;
    acceptable_inited = 1;
}

static char hex[17] = "0123456789abcdef";
static char buffer[1024];

/* hex_encode:  substitute some chars to %## (hex) 
 *              -- for some "imperfect gateways" :-)
 */

char *
hex_encode (in)
char	*in;
{
	char *out = buffer;
	
	bzero(out, 1024);
	if (acceptable_inited == 0)
        	init_acceptable();
	while (*in) {
        	if (acceptable[(unsigned int)*in] == 1) {
               		*out++ = *in++;
        	} else {
               	*out++ = '%';
               	*out++ = hex[(unsigned int)*in >> 4];
               	*out++ = hex[(unsigned int)*in & 15];
               	in++;
        	}
	}
     	*out = '\0';

 	/* fprintf( stderr, "returning - esc: %s.\n", buffer); */
	return (buffer);
}
char *
hex_decode (in)
char *  in;
{
        char b, c;
	int q = 0;
        char *out = in;

        while (*in) {
		if (*in == '?')		/* start search */
			q = 1;
                if (*in == '%') {	/* Hex escape */
                        in++;
                        c = *in++;
                        b = from_hex(c);
                        c = *in++;
                        if (!c) break;
                        *out++ = (b<<4) + from_hex(c);
                } else if (q && *in == '+') {
			/* '+' is legal in path, in search it's a ' ' */
			*out++ = ' ';
			in++;
		} else {
                        *out++ = *in++;
                }
        }
        *out = '\0';
        return (out);
}

/* decode in search (for do_modify) */

char *
hex_qdecode (in)
char *  in;
{
        char b, c;
        char *out = in;

        while (*in) {
                if (*in == '%') { 	/* Hex escape */
                        in++;
                        c = *in++;
                        b = from_hex(c);
                        c = *in++;
                        if (!c) break;
                        *out++ = (b<<4) + from_hex(c);
                } else if (*in == '+') { /* we are in search, so: '+' -> ' ' */
                        *out++ = ' ';
                        in++;
                } else {
                        *out++ = *in++;
                }
        }
        *out = '\0';
        return (out);
}

char *
html_encode (in)
char	*in;
{
	char *out = buffer;
	
	/* bzero(out, 1024); */
	if (acceptable_inited == 0)
        	init_acceptable();
	while (*in) {
        	if (acceptable[(unsigned int)*in] == 1) {
               		*out++ = *in++;
        	} else {
               	*out++ = '&';
		*out++ = '#';
               	*out++ = ((unsigned int)*in / 100) + '0';
               	*out++ = (((unsigned int)*in % 100) / 10) + '0';
		*out++ = (((unsigned int)*in % 100) % 10) + '0';
		*out++ = ';';
               	in++;
        	}
	}
     	*out = '\0';

 	/* fprintf( stderr, "returning - esc: %s.\n", buffer); */
	return (buffer);
}

char *
form_encode (in)
char	*in;
{
	char *out = buffer;
	
	/* bzero(out, 1024); */
	while (*in) {
        	if (*in == '"' || *in == '>' ) {
			*out++ = '\\';
		}
               	*out++ = *in++;
	}
     	*out = '\0';

 	/* fprintf( stderr, "returning - esc: %s.\n", buffer); */
	return (buffer);
}

#ifdef HREFDN
/* Make UFN components to hypertext links */

char *
href_dn (dn, action)
char    *dn;
int     action;
{
        char **s, **t, 
		fhdn[2048],	/* full hypertexted DN, return value */
		thdn[1024],	/* temp */
		hrdn[1024];	/* temp, hypertexted RDN */
        int  cnt, i, j;
	
	fhdn[0] = '\0';
        thdn[0] = '\0';
	hrdn[0] = '\0';
        s = ldap_explode_dn( dn, 1 );
        t = ldap_explode_dn( dn, 0 );
 /*       strcpy(rdn, ldap_friendly_name( friendlyfile, s[0], &fm )); */
        for (cnt=0; t[cnt]; cnt++);
        for (i=0; s[i]; i++) {
		strcpy(hrdn, t[i]);
                for (j=i+1; j < cnt ; j++) {
                        strcat(hrdn, ", ");
                        strcat(hrdn,t[j]);
                }
		
                sprintf( thdn, "<A HREF=\"/%s%s\">%s</A>%s ",
                        action || i > 0 ? "M" : "R", hex_encode(hrdn),
                        ldap_friendly_name( friendlyfile, s[i], &fm ),
                        i == cnt - 1 ? "" : ",");
		strcat(fhdn, thdn);
                hrdn[0] = '\0';
		thdn[0] = '\0';
        }
        ldap_value_free(s);
        ldap_value_free(t);
        return fhdn;
}
#endif HREFDN

/* Make "Move upwards" Header */

make_header (fp, dn, action)
FILE	*fp;
char    *dn;
int     action;
{
        char **s, **t, hrdn[1024];
        int  cnt, i, j;
	hrdn[0] = '\0';
	if ( strlen(dn) == 0)		/* the root */
		return;
        s = ldap_explode_dn( dn, 1 );
        t = ldap_explode_dn( dn, 0 );
	fprintf( fp, "Move upwards to <MENU>\n<LI> <A HREF=\"/M\">%s</A>\n", ROOT);
        for (cnt=0; t[cnt]; cnt++);
        for (i = cnt - 1; i > 0 && s[i]; i--) {
		strcpy(hrdn, t[i]);
                for (j = i + 1; j < cnt ; j++) {
                        strcat(hrdn, ", ");
                        strcat(hrdn,t[j]);
                }
		
                fprintf( fp, "<LI> <A HREF=\"/M%s\">%s</A>\n", hex_encode(hrdn),
			 i == cnt - 1 ? ldap_friendly_name( friendlyfile, s[i], &fm ) : s[i]);
                hrdn[0] = '\0';
        }
	fprintf( fp, "</MENU><HR>\n");
        ldap_value_free(s);
        ldap_value_free(t);
}

/* gtime(): the inverse of localtime().
	This routine was supplied by Mike Accetta at CMU many years ago.
 */

int	dmsize[] = {
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

#define	dysize(y)	\
	(((y) % 4) ? 365 : (((y) % 100) ? 366 : (((y) % 400) ? 365 : 366)))

#define	YEAR(y)		((y) >= 100 ? (y) : (y) + 1900)

static
time_t	gtime (tm)
register struct tm *tm;
{
    register int    i,
                    sec,
                    mins,
                    hour,
                    mday,
                    mon,
                    year;
    register long   result;

    if ((sec = tm -> tm_sec) < 0 || sec > 59
	    || (mins = tm -> tm_min) < 0 || mins > 59
	    || (hour = tm -> tm_hour) < 0 || hour > 24
	    || (mday = tm -> tm_mday) < 1 || mday > 31
	    || (mon = tm -> tm_mon + 1) < 1 || mon > 12)
	return ((long) -1);
    if (hour == 24) {
	hour = 0;
	mday++;
    }
    year = YEAR (tm -> tm_year);
    result = 0L;
    for (i = 1970; i < year; i++)
	result += dysize (i);
    if (dysize (year) == 366 && mon >= 3)
	result++;
    while (--mon)
	result += dmsize[mon - 1];
    result += mday - 1;
    result = 24 * result + hour;
    result = 60 * result + mins;
    result = 60 * result + sec;
    return result;
}

static char *
format_date (s, format)
char	**s;
char	*format;
{
	char		date[256];
	struct tm       tm, *ntm;
	time_t		t;


	tm.tm_year = 10*((*s)[0] - '0') + ((*s)[1] - '0');
	tm.tm_mon  = 10*((*s)[2] - '0') + ((*s)[3] - '0') -
1;
	tm.tm_mday = 10*((*s)[4] - '0') + ((*s)[5] - '0');
	tm.tm_hour = 10*((*s)[6] - '0') + ((*s)[7] - '0');
	tm.tm_min  = 10*((*s)[8] - '0') + ((*s)[9] - '0');
	tm.tm_sec  = 10*((*s)[10] - '0') + ((*s)[11] - '0');
	tm.tm_isdst = 0;
#ifndef __hpux
	tm.tm_gmtoff = 0;
#endif
	t = gtime(&tm);
	ntm = gmtime(&t);
	strftime(date, 256, format, ntm);
	return (strdup(date));
}

static char *
friendly_dn (dn)
char	*dn;
{
	char	fufn[1024], *ufn, *cp;
	
	if (strlen(dn) == 0)
		return ROOT;
        ufn = clean_ufn(ldap_dn2ufn(dn));
				/* now make friendly name for the toplevel */
	if ( cp = strrchr(ufn,',') ) {
		*cp++ = '\0';
		while (*cp == ' '  ) cp++;
		strcpy( fufn, ufn);
		strcat( fufn, ", ");
		strcat( fufn, ldap_friendly_name( friendlyfile, cp, &fm ));
	} else {
		strcpy( fufn, ldap_friendly_name( friendlyfile, ufn, &fm ));
	}
	return (strdup(fufn));
}

static char *
clean_ufn(ufn)
char	*ufn;
{
	char *cp, *cp2, *new;

	/* We make sure to eliminate any unknown attributes in the UFN! */
	if (ufn == NULL)
		return NULL;
	new = strdup(ufn);
	while ( (cp2 = strchr(new, '=')) != NULL ) {
		cp = cp2;
		while ( *cp != ' ' && cp > new ) cp--;
		if  ( cp > new ) {
			*++cp = '\0';
			strcat(new, ++cp2);
		} else
			new = cp2;
	}
	if ( strlen(new) > 0 )
		free(ufn);
	return(new);
}

static 
sort_result( ld, res, dn )
LDAP    *ld;
LDAPMessage     *res;
char    *dn;
{
	LDAPMessage	*e;
	char	**s, **oc;
	char	*dn2, *rdn, *doc, *foc;
	char	*ufn, *ufn2, *sortstring, *cp;
	char   **sn, href[1024], *temp, **val;
	int spaces = 0, counter = 0, baselen, iscountry, isalias = 0;
	
	if ( strlen(dn) > 0 )
		ufn = friendly_dn(dn);
	else
		ufn = NULL;
	baselen = ufn ? strlen(ufn) : 0;

	for ( e = ldap_first_entry( ld, res ); e != NULL && counter < MAXCOUNT;
		e = ldap_next_entry( ld, e ) ) {

		dn2 = ldap_get_dn( ld, e );
		s = ldap_explode_dn( dn2, 1 );
		ufn2 = friendly_dn(dn2);
		oc = ldap_get_values( ld, e, "objectClass" );
		doc = pick_oc( oc );
		foc = strdup(ldap_friendly_name(friendlyfile, doc, &fm));
		iscountry = strcasecmp( doc, "country" );
		if ( iscountry == 0 ) {
			rdn = ldap_friendly_name( friendlyfile, s[0], &fm );
			sortstring = ufn2 = ldap_friendly_name( friendlyfile, ufn2, &fm );
			sn = NULL;
		} else
			rdn = s[0];
		if ( rdn == NULL )
			rdn = s[0];
		if ( strncasecmp( rdn, "{ASN}", 5 ) == 0 ) {
			/* ASNs are not something we want to display! */
			free( dn2 );
			ldap_value_free( s );
			ldap_value_free( oc );
		    	continue;
		}
		if ( baselen > 0 )
			ufn2[strlen(ufn2)-baselen-2] = '\0';
		/* if (debug) fprintf(stderr, "ufn: %s\n", ufn2); */
		if (val = ldap_get_values( ld, e, "aliasedObjectName")) {
				/* entry is alias */
			free(dn2);
			dn2 = strdup(*val);
			ldap_value_free(val);
			isalias = 1;
		} else
			isalias = 0;
		
		if ( iscountry != 0 ) {	/* not a country */
			sn = ldap_get_values( ld, e, "sn");
			sortstring = strdup(ufn2);
			if ( *sortstring == '"' ) 
				cp = strchr(sortstring+1, '"');
			else
				cp = sortstring;
			if (cp = strchr(cp,','))
				*cp = '\0';
			/* DNs may have components in '"', ignored  when sorting */
			if ( *sortstring == '"' )
				sortstring++;
			if ( sn ) {
				cp = *sn;
				while ( cp = strchr(cp,' ') ) {
					cp ++;
					spaces ++;
				}
			}
#ifdef RUTGERS
			if ( cp = strchr(sortstring,'+') ) {
				cp --;
				*cp = '\0';
			}
#endif
			while ( spaces > 0 ) {
				if ( cp = strrchr(sortstring,' ')) {
					*cp = '\0';
					spaces --;
				} else
					break;
			}
		}
#ifdef DEREFALIAS
		if ( isalias == 1 ) {
			/* if alias, look for objectClass of real entry */
			/* -- expensive ?! */
			deref(ld, dn2, oc);
		}
#endif
		if ( isnonleaf( ld, oc, dn2 ) )
			sprintf( href, "<LI><A HREF=\"/M%s\">%s</A> (%s)\n",
			hex_encode(dn2), ufn2, foc);
		else
			sprintf( href, "<LI><A HREF=\"/R%s\">%s</A> (%s)\n",
			hex_encode(dn2), ufn2, foc);
		if ( sn ) 
			temp = (char *) calloc(1,strlen(foc)+strlen(*sn)+strlen(sortstring)+1);
		else 
			temp = (char *) calloc(1,strlen(foc)+strlen(sortstring)+1);
		strcpy(temp, foc);
		if ( sn )
			strcat(temp, *sn);
		strcat(temp,sortstring);
		if (!dnlist[counter] )
			dnlist[counter] = (struct dncompare *) calloc(1,sizeof(struct dncompare));
		dnlist[counter]->string = temp;
		dnlist[counter++]->href = strdup(href);
		free( dn2 );
		free( foc );
		ldap_value_free( s );
		ldap_value_free( oc );
	}
	qsort(dnlist,counter,sizeof(int), compare);
	return counter;
}

#ifdef DEREFALIAS
static void
deref(ld, dn, oc )
LDAP  *ld;
char *dn;
char **oc;
{
	LDAPMessage *e, *res;
	struct timeval  timeout;
	static char   *retattrs[] = { "objectClass" ,0 };
	
	timeout.tv_sec = TIMEOUT;
	timeout.tv_usec = 0;
	if ( ldap_search_st(ld, dn, LDAP_SCOPE_BASE,"objectClass=*", retattrs, 0, &timeout, &res) == LDAP_SUCCESS ) {
		e = ldap_first_entry(ld, res);
		ldap_value_free(oc);
		oc = ldap_get_values( ld, e, "objectClass");
		ldap_msgfree(res);
       	}
}
#endif	
